# -*- coding: utf-8 -*-

"""
DE - Diagnostic Efficiency
"""

from ._version import get_versions

__title__ = "de"
__author__ = "Robin Schwemmle"
__license__ = "GNU GPLv3"
__version__ = get_versions()["version"]
del get_versions
